#********************************************************************************
#  Copyright (c) 2018 Edgeworx, Inc.
#
#  This program and the accompanying materials are made available under the
#  terms of the Eclipse Public License v. 2.0 which is available at
#  http://www.eclipse.org/legal/epl-2.0
#
#  SPDX-License-Identifier: EPL-2.0
#********************************************************************************

class IoFogException(Exception):
    def __init__(self, *args, **kwargs):
        super(IoFogException, self).__init__(*args, **kwargs)


class IoFogHttpException(IoFogException):
    def __init__(self, code, message):
        self.code = code
        self.message = message

    def __str__(self):
        return 'Error code: {}, reason: {}'.format(self.code, self.message)









#********************************************************************************
#  Copyright (c) 2018 Edgeworx, Inc.
#
#  This program and the accompanying materials are made available under the
#  terms of the Eclipse Public License v. 2.0 which is available at
#  http://www.eclipse.org/legal/epl-2.0
#
#  SPDX-License-Identifier: EPL-2.0
#********************************************************************************

IOFOG = 'iofog'

PORT_IOFOG = 54321
SELFNAME = 'SELFNAME'
SSL = 'SSL'
SSL_DEFAULT = False
HOST_DEFAULT = '127.0.0.1'
IOFOG_LOGGER = 'iofog_logger'

URL_GET_CONFIG = '/v2/config/get'
URL_GET_NEXT_MESSAGES = '/v2/messages/next'
URL_GET_PUBLISHERS_MESSAGES = '/v2/messages/query'
URL_POST_MESSAGE = '/v2/messages/new'
URL_GET_CONTROL_WS = '/v2/control/socket/id/'
URL_GET_MESSAGE_WS = '/v2/message/socket/id/'

APPLICATION_JSON = 'application/json'
HTTP = 'http'
HTTPS = 'https'
WS = 'ws'
WSS = 'wss'

CODE_ACK = 0xB
CODE_CONTROL_SIGNAL = 0xC
CODE_MSG = 0xD
CODE_RECEIPT = 0xE

WS_ATTEMPT_LIMIT = 5
IO_MESSAGE_VERSION = 4

WS_CONNECT_TIMEOUT = 1
PING_INTERVAL_SECONDS = 5
CODE_BAD_REQUEST = 400

ID = 'id'
TAG = 'tag'
GROUP_ID = 'groupid'
VERSION = 'version'
SEQUENCE_NUMBER = 'sequencenumber'
SEQUENCE_TOTAL = 'sequencetotal'
PRIORITY = 'priority'
TIMESTAMP = 'timestamp'
PUBLISHER = 'publisher'
AUTH_ID = 'authid'
AUTH_GROUP = 'authgroup'
CHAIN_POSITION = 'chainposition'
HASH = 'hash'
PREVIOUS_HASH = 'previoushash'
NONCE = 'nonce'
DIFFICULTY_TARGET = 'difficultytarget'
INFO_TYPE = 'infotype'
INFO_FORMAT = 'infoformat'
CONTEXT_DATA = 'contextdata'
CONTENT_DATA = 'contentdata'

TIME_FRAME_START = 'timeframestart'
TIME_FRAME_END = 'timeframeend'
PUBLISHERS = 'publishers'
CONFIG = 'config'
MESSAGES = 'messages'
STATUS = 'status'
COUNT = 'count'



#********************************************************************************
#  Copyright (c) 2018 Edgeworx, Inc.
#
#  This program and the accompanying materials are made available under the
#  terms of the Eclipse Public License v. 2.0 which is available at
#  http://www.eclipse.org/legal/epl-2.0
#
#  SPDX-License-Identifier: EPL-2.0
#********************************************************************************

import json
import math

try:
    import urllib.request as urllib_request #for python 3
except ImportError:
    import urllib2 as urllib_request # for python 2

from struct import pack
from iofog.microservices.definitions import CODE_MSG


def num_to_bytearray(num):
    if num == 0:
        return bytearray([0]), 1

    num_of_bits = num.bit_length()
    num_of_bytes = int(math.ceil(num_of_bits / 8.0))
    b = bytearray(num_of_bytes)
    shift = 8 * (num_of_bytes - 1)
    for i in range(num_of_bytes):
        b[i] = (num >> shift) & 0xFF
        shift -= 8
    return b, num_of_bytes


def bytearray_to_num(arr):
    num = 0
    shift = 0
    for i in reversed(arr):
        num += i << shift
        shift += 8
    return num


def make_post_request(url, body_type, body):
    req = urllib_request.Request(url, body, {'Content-Type': body_type})
    response = urllib_request.urlopen(req)
    return json.loads(response.read())


def prepare_iomessage_for_sending_via_socket(io_msg):
    msg_bytes = io_msg.to_bytearray()
    package = bytearray([CODE_MSG])
    package.extend(pack('>I', len(msg_bytes)))
    return package + msg_bytes



#********************************************************************************
#  Copyright (c) 2018 Edgeworx, Inc.
#
#  This program and the accompanying materials are made available under the
#  terms of the Eclipse Public License v. 2.0 which is available at
#  http://www.eclipse.org/legal/epl-2.0
#
#  SPDX-License-Identifier: EPL-2.0
#********************************************************************************

class IoFogControlWsListener:
    def __init__(self):
        pass

    def on_control_signal(self):
        pass


class IoFogMessageWsListener:
    def __init__(self):
        pass

    def on_message(self, io_msg):
        pass

    def on_receipt(self, message_id, timestamp):
        pass



#********************************************************************************
#  Copyright (c) 2018 Edgeworx, Inc.
#
#  This program and the accompanying materials are made available under the
#  terms of the Eclipse Public License v. 2.0 which is available at
#  http://www.eclipse.org/legal/epl-2.0
#
#  SPDX-License-Identifier: EPL-2.0
#********************************************************************************

import json_logging, logging, sys
import datetime
import traceback
import json
import os

class CustomJSONLog(logging.Formatter):
    """
    Customized logger
    """

    def format(self, record):
        json_log_object = {"timestamp": datetime.datetime.utcnow().isoformat(),
                           "level": record.levelname.lower(),
                           "message": record.getMessage(),
                           }
        return json.dumps(json_log_object)

json_logging.init_non_web(custom_formatter=CustomJSONLog, enable_json=True)

class BaseLogger():

    def __init__(self, name):
        self.logger = logging.getLogger(name)
        self.logger.setLevel(logging.DEBUG)
    
    def info(self, msg):
        self.logger.info(msg)

    def debug(self, msg):
        self.logger.debug(msg)

    def warn(self, msg):
        self.logger.warn(msg)

    def error(self, msg):
        self.logger.error(msg)

class Logger(BaseLogger):

    def __init__(self, name):
        super().__init__(name)
        # Create log file
        log_file = "/var/log/iofog-microservices/{}.log".format(name)
        if not os.path.exists(log_file):
            with open(log_file, 'w'):
                pass
        # Register log file
        self.logger.addHandler(logging.FileHandler(log_file))

if __name__ == "__main__":
    log = Logger("werserge")
    log.info("hi")



#********************************************************************************
#  Copyright (c) 2018 Edgeworx, Inc.
#
#  This program and the accompanying materials are made available under the
#  terms of the Eclipse Public License v. 2.0 which is available at
#  http://www.eclipse.org/legal/epl-2.0
#
#  SPDX-License-Identifier: EPL-2.0
#********************************************************************************

import argparse
import logging
import os

import subprocess

from iofog.microservices.httpclient import IoFogHttpClient
from iofog.microservices.definitions import *
from iofog.microservices.wsclient import IoFogControlWsClient, IoFogMessageWsClient
from iofog.microservices.listener import *
from iofog.microservices.exception import *

parser = argparse.ArgumentParser()
parser.add_argument("-l", "--log", dest="logLevel", choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
                    help="Set the logging level", default='INFO')
args = parser.parse_args()


class IoFogClient:
    def __init__(self, id=None, ssl=None, host=None, port=None):
        self.logger = logging.getLogger(IOFOG_LOGGER)
        self.logger.setLevel(args.logLevel)
        ch = logging.StreamHandler()
        ch.setLevel(args.logLevel)
        formatter = logging.Formatter(
            '%(levelname)5s [%(asctime)-15s] %(module)10s - <Thread: %(threadName)15s> - %(message)s')
        ch.setFormatter(formatter)
        self.logger.addHandler(ch)
        if id:
            self.id = id
        else:
            self.id = os.getenv(SELFNAME)
            if not self.id:
                raise IoFogException('Cannot create client with empty id: ' +
                                     SELFNAME + ' environment variable is not set')

        if ssl:
            self.ssl = ssl
        else:
            self.ssl = os.getenv(SSL)
            if not self.ssl:
                self.logger.info('Empty or malformed ' + SSL
                                 + ' environment variable. Using default value of ' + str(SSL_DEFAULT))
                self.ssl = SSL_DEFAULT

        if host:
            self.host = host
        else:
            self.host = IOFOG
            with open(os.devnull, 'w') as FNULL:
                resp = subprocess.call(['ping', '-c', '3', self.host], stdout=FNULL, stderr=FNULL)
            if resp:
                self.logger.info('Host ' + IOFOG + ' is unreachable. Switching to ' + HOST_DEFAULT)
                self.host = HOST_DEFAULT

        self.port = port if port else PORT_IOFOG

        self.message_ws_client = None
        self.control_ws_client = None
        self.http_client = IoFogHttpClient(self.id, self.ssl, self.host, self.port)

    def establish_message_ws_connection(self, listener):
        if listener is not None and not isinstance(listener, IoFogMessageWsListener):
            raise IoFogException('Invalid listener instance')
        if self.message_ws_client:
            raise IoFogException('Connection has been already established')
        self.message_ws_client = IoFogMessageWsClient(self.id, self.ssl, self.host,
                                                      self.port, URL_GET_MESSAGE_WS, listener)
        self.message_ws_client.connect()

    def establish_control_ws_connection(self, listener):
        if listener is not None and not isinstance(listener, IoFogControlWsListener):
            raise IoFogException('Invalid listener instance')
        if self.control_ws_client:
            raise IoFogException('Connection has been already established')
        self.control_ws_client = IoFogControlWsClient(self.id, self.ssl, self.host,
                                                      self.port, URL_GET_CONTROL_WS, listener)
        self.control_ws_client.connect()

    def get_config(self):
        try:
            return self.http_client.get_config()
        except Exception as e:
            raise IoFogException(e)

    def get_next_messages(self):
        try:
            return self.http_client.get_next_messages()
        except Exception as e:
            raise IoFogException(e)

    def get_next_messages_from_publishers_within_timeframe(self, query):
        assert TIME_FRAME_START in query and TIME_FRAME_END in query and PUBLISHERS in query, 'Wrong query parameters'
        query[ID] = self.id
        try:
            return self.http_client.get_next_messages_from_publishers_within_timeframe(query)
        except Exception as e:
            raise IoFogException(e)

    def post_message(self, io_msg):
        if not io_msg.version:
            io_msg.version = IO_MESSAGE_VERSION
        io_msg.publisher = self.id
        try:
            return self.http_client.post_message(io_msg)
        except Exception as e:
            raise IoFogException(e)

    def post_message_via_socket(self, io_msg):
        if not self.message_ws_client:
            raise IoFogException('Establish message websocket connection first!')
        while not self.message_ws_client.is_open:
            pass
        io_msg.id = ''
        io_msg.timestamp = 0
        io_msg.publisher = self.id
        if io_msg.version == 0:
            io_msg.version = IO_MESSAGE_VERSION
        try:
            self.message_ws_client.send_message(io_msg)
        except Exception as e:
            raise IoFogException(e)



#********************************************************************************
#  Copyright (c) 2018 Edgeworx, Inc.
#
#  This program and the accompanying materials are made available under the
#  terms of the Eclipse Public License v. 2.0 which is available at
#  http://www.eclipse.org/legal/epl-2.0
#
#  SPDX-License-Identifier: EPL-2.0
#********************************************************************************

import json

from struct import pack, unpack
import sys

from iofog.microservices.util import *
import base64

from iofog.microservices.definitions import *


# todo check fields == None?
class IoMessage:
    def __init__(self):
        self.id = ''
        self.tag = ''
        self.groupid = ''
        self.version = IO_MESSAGE_VERSION
        self.sequencenumber = 0
        self.sequencetotal = 0
        self.priority = 0
        self.timestamp = 0
        self.publisher = ''
        self.authid = ''
        self.authgroup = ''
        self.chainposition = 0
        self.hash = ''
        self.previoushash = ''
        self.nonce = ''
        self.difficultytarget = 0
        self.infotype = ''
        self.infoformat = ''
        self.contextdata = bytearray()
        self.contentdata = bytearray()

    def to_bytearray(self):
        if self.version != IO_MESSAGE_VERSION:
            raise Exception('Incompatible IoMessage version')

        header = bytearray()
        body = bytearray()
        # version
        header.extend(pack('>H', self.version))
        # id
        header.extend(pack('>B', len(self.id)))
        body.extend(self.id.encode())
        # tag
        header.extend(pack('>H', len(self.tag)))
        body.extend(self.tag.encode())
        # group id
        header.extend(pack('>B', len(self.groupid)))
        body.extend(self.groupid.encode())
        # sequence number
        byte_num, byte_num_len = num_to_bytearray(self.sequencenumber)
        header.extend(pack('>B', byte_num_len))
        body.extend(byte_num.encode())
        # sequence total
        byte_num, byte_num_len = num_to_bytearray(self.sequencetotal)
        header.extend(pack('>B', byte_num_len))
        body.extend(byte_num.encode())
        # priority
        byte_num, byte_num_len = num_to_bytearray(self.priority)
        header.extend(pack('>B', byte_num_len))
        body.extend(byte_num.encode())
        # timestamp
        byte_num, byte_num_len = num_to_bytearray(self.timestamp)
        header.extend(pack('>B', byte_num_len))
        body.extend(byte_num.encode())
        # publisher
        header.extend(pack('>B', len(self.publisher)))
        body.extend(self.publisher.encode())
        # auth id
        header.extend(pack('>H', len(self.authid)))
        body.extend(self.authid.encode())
        # auth group
        header.extend(pack('>H', len(self.authgroup)))
        body.extend(self.authgroup.encode())
        # chain position
        byte_num, byte_num_len = num_to_bytearray(self.chainposition)
        header.extend(pack('>B', byte_num_len))
        body.extend(byte_num.encode())
        # hash
        header.extend(pack('>H', len(self.hash)))
        body.extend(self.hash.encode())
        # previous hash
        header.extend(pack('>H', len(self.previoushash)))
        body.extend(self.previoushash.encode())
        # nonce
        header.extend(pack('>H', len(self.nonce)))
        body.extend(self.nonce.encode())
        # difficulty target
        byte_num, byte_num_len = num_to_bytearray(self.difficultytarget)
        header.extend(pack('>B', byte_num_len))
        body.extend(byte_num.encode())
        # info type
        header.extend(pack('>B', len(self.infotype)))
        body.extend(self.infotype.encode())
        # info format
        header.extend(pack('>B', len(self.infoformat)))
        body.extend(self.infoformat.encode())
        # context data
        header.extend(pack('>I', len(self.contextdata)))
        body.extend(self.contextdata.encode())
        # content data
        header.extend(pack('>I', len(self.contentdata)))
        body.extend(self.contentdata.encode())

        return header + body

    @staticmethod
    def from_bytearray(bytes):
        msg = IoMessage()
        msg.version = unpack('>H', bytes[:2])[0]
        if msg.version != IO_MESSAGE_VERSION:
            raise Exception('Incompatible IoMessage version')

        data_offset = 33
        # id
        field_len = bytes[2]
        msg.id = str(bytes[data_offset: data_offset + field_len])
        data_offset += field_len
        # tag
        field_len = unpack('>H', bytes[3:5])[0]
        msg.tag = str(bytes[data_offset: data_offset + field_len])
        data_offset += field_len
        # group id
        field_len = bytes[5]
        msg.groupid = str(bytes[data_offset: data_offset + field_len])
        data_offset += field_len
        # sequence number
        field_len = bytes[6]
        msg.sequencenumber = bytearray_to_num(bytes[data_offset: data_offset + field_len])
        data_offset += field_len
        # sequence total
        field_len = bytes[7]
        msg.sequencetotal = bytearray_to_num(bytes[data_offset: data_offset + field_len])
        data_offset += field_len
        # priority
        field_len = bytes[8]
        msg.priority = bytearray_to_num(bytes[data_offset: data_offset + field_len])
        data_offset += field_len
        # timestamp
        field_len = bytes[9]
        msg.timestamp = bytearray_to_num(bytes[data_offset: data_offset + field_len])
        data_offset += field_len
        # publisher
        field_len = bytes[10]
        msg.publisher = str(bytes[data_offset: data_offset + field_len])
        data_offset += field_len
        # auth id
        field_len = unpack('>H', bytes[11:13])[0]
        msg.authid = str(bytes[data_offset: data_offset + field_len])
        data_offset += field_len
        # auth group
        field_len = unpack('>H', bytes[13:15])[0]
        msg.authgroup = str(bytes[data_offset: data_offset + field_len])
        data_offset += field_len
        # chain position
        field_len = bytes[15]
        msg.chainposition = bytearray_to_num(bytes[data_offset: data_offset + field_len])
        data_offset += field_len
        # hash
        field_len = unpack('>H', bytes[16:18])[0]
        msg.hash = str(bytes[data_offset: data_offset + field_len])
        data_offset += field_len
        # previous hash
        field_len = unpack('>H', bytes[18:20])[0]
        msg.previoushash = str(bytes[data_offset: data_offset + field_len])
        data_offset += field_len
        # nonce
        field_len = unpack('>H', bytes[20:22])[0]
        msg.nonce = str(bytes[data_offset: data_offset + field_len])
        data_offset += field_len
        # difficulty target
        field_len = bytes[22]
        msg.difficultytarget = bytearray_to_num(bytes[data_offset: data_offset + field_len])
        data_offset += field_len
        # info type
        field_len = bytes[23]
        msg.infotype = str(bytes[data_offset: data_offset + field_len])
        data_offset += field_len
        # info format
        field_len = bytes[24]
        msg.infoformat = str(bytes[data_offset: data_offset + field_len])
        data_offset += field_len
        # context data
        field_len = unpack('>I', bytes[25:29])[0]
        msg.contextdata = bytes[data_offset: data_offset + field_len]
        data_offset += field_len
        # content data
        field_len = unpack('>I', bytes[29:33])[0]
        msg.contentdata = bytes[data_offset: data_offset + field_len]
        return msg

    @staticmethod
    def from_json(json_msg):
        if isinstance(json_msg, str):
            json_msg = json.loads(json_msg)
        msg = IoMessage()
        msg.id = json_msg.get(ID, '')
        msg.tag = json_msg.get(TAG, '')
        msg.groupid = json_msg.get(GROUP_ID, '')
        msg.version = json_msg.get(VERSION, 0)
        msg.timestamp = json_msg.get(TIMESTAMP, 0)
        msg.sequencenumber = json_msg.get(SEQUENCE_NUMBER, 0)
        msg.sequencetotal = json_msg.get(SEQUENCE_TOTAL, 0)
        msg.priority = json_msg.get(PRIORITY, 0)
        msg.publisher = json_msg.get(PUBLISHER, '')
        msg.authid = json_msg.get(AUTH_ID, '')
        msg.authgroup = json_msg.get(AUTH_GROUP, '')
        msg.chainposition = json_msg.get(CHAIN_POSITION, 0)
        msg.hash = json_msg.get(HASH, '')
        msg.previoushash = json_msg.get(PREVIOUS_HASH, '')
        msg.nonce = json_msg.get(NONCE, '')
        msg.difficultytarget = json_msg.get(DIFFICULTY_TARGET, 0)
        msg.infotype = json_msg.get(INFO_TYPE, '')
        msg.infoformat = json_msg.get(INFO_FORMAT, '')
        contextdata = json_msg.get(CONTEXT_DATA, '')
        contextdata = base64.b64decode(contextdata)
        msg.contextdata = str.encode(contextdata)
        contentdata = json_msg.get(CONTENT_DATA, '')
        contentdata = base64.b64decode(contentdata)
        msg.contentdata = str.encode(contentdata)
        return msg

    def to_json(self):
        json_msg = {
            VERSION: self.version,
            ID: self.id,
            TAG: self.tag,
            GROUP_ID: self.groupid,
            TIMESTAMP: self.timestamp,
            SEQUENCE_NUMBER: self.sequencenumber,
            SEQUENCE_TOTAL: self.sequencetotal,
            PRIORITY: self.priority,
            PUBLISHER: self.publisher,
            AUTH_ID: self.authid,
            AUTH_GROUP: self.authgroup,
            CHAIN_POSITION: self.chainposition,
            HASH: self.hash,
            PREVIOUS_HASH: self.previoushash,
            NONCE: self.nonce,
            DIFFICULTY_TARGET: self.difficultytarget,
            INFO_TYPE: self.infotype,
            INFO_FORMAT: self.infoformat,
            CONTEXT_DATA: base64.b64encode(self.contextdata),
            CONTENT_DATA: base64.b64encode(self.contentdata)
        }
        return json.dumps(json_msg)



#********************************************************************************
#  Copyright (c) 2018 Edgeworx, Inc.
#
#  This program and the accompanying materials are made available under the
#  terms of the Eclipse Public License v. 2.0 which is available at
#  http://www.eclipse.org/legal/epl-2.0
#
#  SPDX-License-Identifier: EPL-2.0
#********************************************************************************

import logging
import threading

import time

import iofog.microservices.util as util
from iofog.microservices.iomessage import IoMessage
from iofog.microservices.definitions import *
from ws4py.client.threadedclient import WebSocketClient
from ws4py.framing import OPCODE_PONG


class IoFogWsClient(WebSocketClient):
    def __init__(self, container_id, ssl, host, port, url, listener):
        protocol_ws = WS
        if ssl:
            protocol_ws = WSS

        self.logger = logging.getLogger(IOFOG_LOGGER)
        self.url_base_ws = '{}://{}:{}'.format(protocol_ws, host, port)
        self.url_ws = self.url_base_ws + url + container_id
        self.wsAttempt = 0
        self.listener = listener
        self.worker = None
        self.is_open = False
        super(IoFogWsClient, self).__init__(self.url_ws)

    def _reconnect(self, code=None, reason=None):
        self.is_open = False
        self.logger.info('WebSocket connection is closed{}{}. Reconnecting...'.format(
            '' if not code else ' with code {}'.format(code),
            '' if not reason else '({})'.format(reason)))
        sleep_time = 1 << self.wsAttempt * WS_CONNECT_TIMEOUT
        if self.wsAttempt < WS_ATTEMPT_LIMIT:
            self.wsAttempt += 1
        try:
            time.sleep(sleep_time)
        except KeyboardInterrupt:
            pass

    def process(self, bytes):
        s = self.stream

        if not bytes and self.reading_buffer_size > 0:
            return False

        from ws4py.websocket import DEFAULT_READING_SIZE
        self.reading_buffer_size = s.parser.send(bytes) or DEFAULT_READING_SIZE

        if s.closing is not None:
            self.logger.debug("Closing message received (%d) '%s'" % (s.closing.code, s.closing.reason))
            if not self.server_terminated:
                self.close(s.closing.code, s.closing.reason)
            else:
                self.client_terminated = True
            return False

        if s.errors:
            for error in s.errors:
                self.logger.debug("Error message received (%d) '%s'" % (error.code, error.reason))
                self.close(error.code, error.reason)
            s.errors = []
            return False

        if s.has_message:
            self.received_message(s.message)
            if s.message is not None:
                s.message.data = None
                s.message = None
            return True

        if s.pings:
            for _ in s.pings:
                self.logger.debug('Got ping from iofog')
                self._write(s.pong(bytearray([OPCODE_PONG])))
            s.pings = []

        if s.pongs:
            for pong in s.pongs:
                self.ponged(pong)
            s.pongs = []

        return True

    def opened(self):
        self.wsAttempt = 0
        self.is_open = True

    def connect(self):
        self.logger.debug('Starting connect')
        self.worker = threading.Thread(target=self._serve, name='WS Server')
        self.worker.start()

    def _serve(self):
        self.logger.debug('Starting serving')
        while True:
            try:
                super(IoFogWsClient, self).connect()
            except Exception as e:
                self._reconnect(reason=e.message)
                continue
            self.run_forever()
            self.logger.debug('Loop exited')
            super(IoFogWsClient, self).__init__(self.url_ws)
            self._reconnect(reason='Connection terminated')


class IoFogControlWsClient(IoFogWsClient):
    def __init__(self, container_id, ssl, host, port, url, listener):
        super(IoFogControlWsClient, self).__init__(container_id, ssl, host, port, url, listener)

    def received_message(self, message):
        opt_code = bytearray(message.data)[0]
        if opt_code == CODE_CONTROL_SIGNAL:
            self.logger.debug('Received control')
            self.listener.on_control_signal()
            self.send(bytearray([CODE_ACK]), binary=True)


class IoFogMessageWsClient(IoFogWsClient):
    def __init__(self, container_id, ssl, host, port, url, listener):
        super(IoFogMessageWsClient, self).__init__(container_id, ssl, host, port, url, listener)

    def received_message(self, message):
        data = bytearray(message.data)
        opt_code = data[0]
        if opt_code == CODE_MSG:
            self.logger.debug('Received message')
            msg_data = data[5:]
            if len(msg_data) == 0:
                return
            msg = IoMessage.from_bytearray(msg_data)
            self.send(bytearray([CODE_ACK]), binary=True)
            self.listener.on_message(msg)
        elif opt_code == CODE_RECEIPT:
            self.logger.debug('Received receipt')
            receipt_data = data[1:]
            if len(receipt_data) == 0:
                return
            id_len = receipt_data[0]
            ts_len = receipt_data[1]
            pos = 2
            message_id = str(receipt_data[pos: pos + id_len])
            pos += id_len
            timestamp = util.bytearray_to_num(receipt_data[pos: pos + ts_len])
            self.send(bytearray([CODE_ACK]), binary=True)
            self.listener.on_receipt(message_id, timestamp)

    def send_message(self, io_msg):
        self.logger.debug('Sending message')
        self.send(util.prepare_iomessage_for_sending_via_socket(io_msg), binary=True)
        self.logger.debug('Sent message')



#********************************************************************************
#  Copyright (c) 2018 Edgeworx, Inc.
#
#  This program and the accompanying materials are made available under the
#  terms of the Eclipse Public License v. 2.0 which is available at
#  http://www.eclipse.org/legal/epl-2.0
#
#  SPDX-License-Identifier: EPL-2.0
#********************************************************************************

import json

try:
    from urllib.error import HTTPError #for python 3
except ImportError:
    from urllib2 import HTTPError #for python 2

from iofog.microservices.definitions import *
from iofog.microservices.util import make_post_request
from iofog.microservices.iomessage import IoMessage
from iofog.microservices.exception import IoFogHttpException


class IoFogHttpClient:
    def __init__(self, container_id, ssl, host, port):
        protocol_rest = HTTP
        if ssl:
            protocol_rest = HTTPS

        self.url_base_rest = "{}://{}:{}".format(protocol_rest, host, port)
        self.url_get_config = self.url_base_rest + URL_GET_CONFIG
        self.url_get_next_messages = self.url_base_rest + URL_GET_NEXT_MESSAGES
        self.url_get_publishers_messages = self.url_base_rest + URL_GET_PUBLISHERS_MESSAGES
        self.url_post_message = self.url_base_rest + URL_POST_MESSAGE
        self.request_body_id = json.dumps({
            ID: container_id
        })

    def get_config(self):
        try:
            config_resp = make_post_request(self.url_get_config, APPLICATION_JSON, self.request_body_id)
        except HTTPError as e:
            raise IoFogHttpException(e.code, e.read())
        return json.loads(config_resp[CONFIG])

    def get_next_messages(self):
        try:
            next_messages_resp = make_post_request(self.url_get_next_messages, APPLICATION_JSON, self.request_body_id)
        except HTTPError as e:
            raise IoFogHttpException(e.code, e.read())
        messages = []
        for json_msg in next_messages_resp[MESSAGES]:
            messages.append(IoMessage.from_json(json_msg))
        return messages

    def get_next_messages_from_publishers_within_timeframe(self, query):
        try:
            next_messages_resp = make_post_request(self.url_get_publishers_messages, APPLICATION_JSON,
                                                   json.dumps(query))
        except HTTPError as e:
            raise IoFogHttpException(e.code, e.read())
        response = {
            TIME_FRAME_START: next_messages_resp[TIME_FRAME_START],
            TIME_FRAME_END: next_messages_resp[TIME_FRAME_END]
        }
        messages = []
        for json_msg in next_messages_resp[MESSAGES]:
            messages.append(IoMessage.from_json(json_msg))
        response[MESSAGES] = messages
        return response

    def post_message(self, io_msg):
        try:
            post_resp = make_post_request(self.url_post_message, APPLICATION_JSON, io_msg.to_json())
            del post_resp[STATUS]
            return post_resp
        except HTTPError as e:
            raise IoFogHttpException(e.code, e.read())



from iofog.rest.controller.request import *

class Client:
    """
    Iofog Controller REST API client.
    """

    def __init__(self, host, port, email, password):
        self.base_path = "http://" + host + ":" + str(port) + "/api/v3"
        self._login(email, password)
    
    def _login(self, email, password):
        url = "{}/user/login".format(self.base_path)
        body = {
            "email": email,
            "password": password
        }
        self.token = request("POST", url, "", body)["accessToken"]
    
    def get_status(self):
        url = "{}/status".format(self.base_path)
        return request("GET", url)

    def create_agent(self, name, host):
        url = "{}/iofog".format(self.base_path)
        body = {
            "name": name,
            "fogType": 0,
            "host": host
        }
        return request("POST", url, self.token, body)["uuid"]

    def _get_provision_key(self, agent_id):
        url = "{}/iofog/{}/provisioning-key".format(self.base_path, agent_id)
        return request("GET", url, self.token)["key"]
    
    def _delete_agent(self, agent_id):
        url = "{}/iofog/{}".format(self.base_path, agent_id)
        request("DELETE", url, self.token)
    
    def _get_agent_id(self, name):
        url = "{}/iofog-list".format(self.base_path)
        resp = request("GET", url, self.token)
        for fog in resp["fogs"]:
            if fog["name"] == name:
                return fog["uuid"]

    def delete_agent(self, name):
        uuid = self._get_agent_id(name)
        if uuid is not None:
            self._delete_agent(uuid)
    
    def get_provision_key(self, agent_name):
        uuid = self._get_agent_id(agent_name)
        if uuid is None:
            raise Exception("Could not get Agent UUID")
        return self._get_provision_key(uuid)

    def upgrade_agent(self, agent_name):
        uuid = self._get_agent_id(agent_name)
        url = "{}/iofog/{}/version/upgrade".format(self.base_path, uuid)
        return request("POST", url, self.token)

    def patch_agent(self, agent_name, config):
        uuid = self._get_agent_id(agent_name)
        url = "{}/iofog/{}".format(self.base_path, uuid)
        return request("PATCH", url, self.token, config)



#********************************************************************************
#  Copyright (c) 2018 Edgeworx, Inc.
#
#  This program and the accompanying materials are made available under the
#  terms of the Eclipse Public License v. 2.0 which is available at
#  http://www.eclipse.org/legal/epl-2.0
#
#  SPDX-License-Identifier: EPL-2.0
#********************************************************************************

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='iofog',
    version='0.0.4',
    project_urls={
        'Documentation': 'https://github.com/eclipse-iofog/iofog-python-sdk/blob/master/README.md',
        'Source': 'https://github.com/eclipse-iofog/iofog-python-sdk.git',
        'Tracker': 'https://github.com/eclipse-iofog/iofog-python-sdk/issues',
        'Eclipse ioFog': 'http://iofog.org'
    },
    packages=setuptools.find_packages(),
    url='https://github.com/eclipse-iofog/iofog-python-sdk',
    license='EPL-2.0',
    author='Eclipse ioFog',
    author_email='edgemaster@iofog.org',
    description='Python SDK for Eclipse ioFog development.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    requires=['ws4py', 'json', 'requests'],
    keywords='ioFog IoT Eclipse fog computing edgeworx',
)



import requests
import json


def request(method, address, auth_token="", body={}):
    switch = {
        "POST": requests.post,
        "GET": requests.get,
        "DELETE": requests.delete,
        "PATCH": requests.patch,
    }
    headers = {}
    data = {}
    if body:
        headers["Content-Type"] = 'application/json'
        data = json.dumps(body, indent=4)

    if auth_token:
        headers["Authorization"] = auth_token
    else:
        headers["cache-control"] = "no-cache"

    response = switch[method](address, data=data, headers=headers, timeout=30)
    try:
        response.raise_for_status()
    except requests.HTTPError as e:
        print(e.response.content)
        raise e
    responseDict = {}
    if response.content:
        responseDict = json.loads(response.content)
    return responseDict